/* Copyright (C) 2007 The Written Word, Inc.
 * Copyright (C) 2008, Simon Josefsson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following disclaimer.
 *
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 *   Neither the name of the copyright holder nor the names
 *   of any other contributors may be used to endorse or
 *   promote products derived from this software without
 *   specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

#include "libssh2_priv.h"
#include "pbkdf/bcrypt_pbkdf.h"

static int
readline(char *line, int line_size, FILE * fp)
{
    size_t len;

    if (!line) {
        return -1;
    }
    if (!fgets(line, line_size, fp)) {
        return -1;
    }

    if (*line) {
        len = strlen(line);
        if (len > 0 && line[len - 1] == '\n') {
            line[len - 1] = '\0';
        }
    }

    if (*line) {
        len = strlen(line);
        if (len > 0 && line[len - 1] == '\r') {
            line[len - 1] = '\0';
        }
    }

    return 0;
}

static int
readline_memory(char *line, size_t line_size,
                const char *filedata, size_t filedata_len,
                size_t *filedata_offset)
{
    size_t off, len;

    off = *filedata_offset;

    for (len = 0; off + len < filedata_len && len < line_size; len++) {
        if (filedata[off + len] == '\n' ||
            filedata[off + len] == '\r') {
                break;
        }
    }

    if (len) {
        memcpy(line, filedata + off, len);
        *filedata_offset += len;
    }

    line[len] = '\0';
    *filedata_offset += 1;

    return 0;
}

#define LINE_SIZE 128

int
_libssh2_pem_parse(LIBSSH2_SESSION * session,
                   const char *headerbegin,
                   const char *headerend,
                   FILE * fp, unsigned char **data, unsigned int *datalen,
                   char **cryptodata)
{
    char line[LINE_SIZE];
    char *b64data = NULL;
    unsigned int b64datalen = 0;
    int encryptedHeader=0;
    int ret;

    do {
        *line = '\0';

        if (readline(line, LINE_SIZE, fp)) {
            return -1;
        }
    }
    while (strcmp(line, headerbegin) != 0);

    *line = '\0';

    do {
        if (*line) {
            char *tmp;
            size_t linelen;

            // detect encryption:
            // Proc-Type: 4,ENCRYPTED
            // DEK-Info: AES-128-CBC,EE7C9BD9FC70CBB481F801D2AA3A55CC
            linelen = strlen(line);
            if (encryptedHeader && strchr(line,':')==NULL)
                encryptedHeader=0;   // end of encryption reached!
            if (strnicmp(line,"DEK-INFO:",9)==0) {
                if (cryptodata) {
                    tmp=line+10;
                    while (tmp[0]==' ')
                        tmp++;
                    *cryptodata = LIBSSH2_ALLOC(session, strlen(tmp)+1);
                    memcpy(*cryptodata, tmp, strlen(tmp)+1);   // 0 terminated!
                }
            } else if (strnicmp(line,"PROC-TYPE:",10)==0) {
                encryptedHeader=1;
                b64datalen=0; // discard anything before!
                tmp=line+10;
                while (tmp[0]==' ')
                    tmp++;
                if (strnicmp(tmp,"4,ENCRYPTED",11)!=0) {   // must be 4, otherwise encryption unknown!
                    ret = -1;
                    goto out;
                }
            } else {
                tmp = LIBSSH2_REALLOC(session, b64data, b64datalen + linelen);
                if (!tmp) {
                    ret = -1;
                    goto out;
                }
                memcpy(tmp + b64datalen, line, linelen);
                b64data = tmp;
                b64datalen += linelen;
            }
        }

        *line = '\0';

        if (readline(line, LINE_SIZE, fp)) {
            ret = -1;
            goto out;
        }
    } while (strcmp(line, headerend) != 0);

    if (!b64data) {
        return -1;
    }

    if (libssh2_base64_decode(session, (char**) data, datalen,
                              b64data, b64datalen)) {
        ret = -1;
        goto out;
    }

    ret = 0;
  out:
    if (b64data) {
        LIBSSH2_FREE(session, b64data);
    }
    return ret;
}

int
_libssh2_pem_parse_memory(LIBSSH2_SESSION * session,
                          const char *headerbegin,
                          const char *headerend,
                          const char *filedata, size_t filedata_len,
                          unsigned char **data, unsigned int *datalen)
{
    char line[LINE_SIZE];
    char *b64data = NULL;
    unsigned int b64datalen = 0;
    size_t off = 0;
    int ret;

    do {
        *line = '\0';

        if (readline_memory(line, LINE_SIZE, filedata, filedata_len, &off)) {
            return -1;
        }
    }
    while (strcmp(line, headerbegin) != 0);

    *line = '\0';

    do {
        if (*line) {
            char *tmp;
            size_t linelen;

            linelen = strlen(line);
            tmp = LIBSSH2_REALLOC(session, b64data, b64datalen + linelen);
            if (!tmp) {
                ret = -1;
                goto out;
            }
            memcpy(tmp + b64datalen, line, linelen);
            b64data = tmp;
            b64datalen += linelen;
        }

        *line = '\0';

        if (readline_memory(line, LINE_SIZE, filedata, filedata_len, &off)) {
            ret = -1;
            goto out;
        }
    } while (strcmp(line, headerend) != 0);

    if (!b64data) {
        return -1;
    }

    if (libssh2_base64_decode(session, (char**) data, datalen,
                              b64data, b64datalen)) {
        ret = -1;
        goto out;
    }

    ret = 0;
  out:
    if (b64data) {
        LIBSSH2_FREE(session, b64data);
    }
    return ret;
}

int
read_asn1_length(const unsigned char *data,
                 unsigned int datalen, unsigned int *len)
{
    unsigned int lenlen;
    int nextpos;

    if (datalen < 1) {
        return -1;
    }
    *len = data[0];

    if (*len >= 0x80) {
        lenlen = *len & 0x7F;
        *len = data[1];
        if (1 + lenlen > datalen) {
            return -1;
        }
        if (lenlen > 1) {
            *len <<= 8;
            *len |= data[2];
        }
    } else {
        lenlen = 0;
    }

    nextpos = 1 + lenlen;
    if (lenlen > 2 || 1 + lenlen + *len > datalen) {
        return -1;
    }

    return nextpos;
}

int
_libssh2_pem_decode_sequence(unsigned char **data, unsigned int *datalen)
{
    unsigned int len;
    int lenlen;

    if (*datalen < 1) {
        return -1;
    }

    if ((*data)[0] != '\x30') {
        return -1;
    }

    (*data)++;
    (*datalen)--;

    lenlen = read_asn1_length(*data, *datalen, &len);
    if (lenlen < 0 || lenlen + len != *datalen) {
        return -1;
    }

    *data += lenlen;
    *datalen -= lenlen;

    return 0;
}

int
_libssh2_pem_decode_integer(unsigned char **data, unsigned int *datalen,
                            unsigned char **i, unsigned int *ilen)
{
    unsigned int len;
    int lenlen;

    if (*datalen < 1) {
        return -1;
    }

    if ((*data)[0] != '\x02') {
        return -1;
    }

    (*data)++;
    (*datalen)--;

    lenlen = read_asn1_length(*data, *datalen, &len);
    if (lenlen < 0 || lenlen + len > *datalen) {
        return -1;
    }

    *data += lenlen;
    *datalen -= lenlen;

    *i = *data;
    *ilen = len;

    *data += len;
    *datalen -= len;

    return 0;
}

unsigned int
_libssh2_pem_read32bit(unsigned char **pbEncoded,int *remain)
{
    unsigned int ret;
    unsigned char *pData=*pbEncoded;
    if (*remain<4)
        return 0xFFFF;
    // data in little endian format
    ret = (((unsigned int)pData[0])<<24) |
           (((unsigned int)pData[1])<<16) |
           (((unsigned int)pData[2])<<8) |
           (((unsigned int)pData[3]));
    *pbEncoded+=4;
    *remain-=4;
    return ret;
}

unsigned int
_libssh2_pem_getByteArrayOffset(unsigned char **pbEncoded,int *remain,unsigned char **pbBuffer,unsigned int *buflen) {
    int len=(int)_libssh2_pem_read32bit(pbEncoded,remain);
    if (len>*remain)
        return -1;
    *pbBuffer=*pbEncoded;
    *buflen=len;
    *pbEncoded+=len;
    *remain-=len;
    return 0;
}

const LIBSSH2_CRYPT_METHOD*
_libssh2_pem_findMethodByName(char* cryptodata)
{
    unsigned int method_len;
    const LIBSSH2_CRYPT_METHOD *method=NULL;
    const LIBSSH2_CRYPT_METHOD **cryptp = libssh2_crypt_methods(); 

    method_len=strlen(cryptodata);
    while (*cryptp) {
        if ((strlen((*cryptp)->name) == method_len) &&
            (strncmp((*cryptp)->name, cryptodata, method_len) == 0)) {
            method=*cryptp;
            break;
        }
        cryptp++;
    }
    if (method==NULL && cryptodata[3]=='-') {
        if (stricmp(cryptodata,"DES-EDE3-CBC")==0) {
            strcpy(cryptodata,"3des-cbc");
            method_len=8;
        } else {
            memmove(cryptodata+3,cryptodata+4,strlen(cryptodata+4));
            method_len--;
            cryptodata[method_len]=0;
        }
        cryptp = libssh2_crypt_methods();
        while (*cryptp) {
            if ((strlen((*cryptp)->name) == method_len) &&
                (strncmp((*cryptp)->name, cryptodata, method_len) == 0)) {
                method=*cryptp;
                break;
            }
            cryptp++; 
        }
    }
    return method;
}

int
_libssh2_pem_decrypt_bcrypt(LIBSSH2_SESSION *session,
                         const char *passphrase,
                         unsigned char **data,
                         unsigned int *datalen)
{
    char cipher[32];  // e.g. aes256-cbc
    const LIBSSH2_CRYPT_METHOD *method=NULL;
    unsigned char *pbEncoded = *data;
    unsigned int cbEncoded = *datalen;
    unsigned char *cipherPtr,*pbOptions,*pbSalt,*pbData,*pbPrivateKey,*pbPublicKey;
    unsigned int cbOptions,cbSalt,cbPublicKey,cbPrivateKey;
    unsigned int cipherLen,rounds,blockcount,blocknr,ret,len;
    unsigned char* keyAndIV;
    unsigned int *testNr1,*testNr2;
    _libssh2_cipher_ctx ctx;

    pbEncoded+=15;   // skip the code name "openssh-key-v1"
    cbEncoded-=15;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&cipherPtr,&cipherLen)!=0)
        return -1;
    if (cipherLen>=32)
        return -1;
    strncpy(cipher,cipherPtr,cipherLen);
    cipher[cipherLen]=0;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&cipherPtr,&cipherLen)!=0)
        return -1;
    if ((cipherLen!=6 || strncmp(cipherPtr,"bcrypt",6)!=0) && strcmp(cipher, "none") != 0)
		return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&pbOptions,&cbOptions)!=0)
        return -1;
    len = _libssh2_pem_read32bit(&pbEncoded,&cbEncoded);
    if (len!=1)
        return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&pbPublicKey,&cbPublicKey)!=0)    // public key
		return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&pbPrivateKey,&cbPrivateKey)!=0)    // private key
        return -1;
	if (strcmp(cipher, "none") == 0) {  // return unencrypted private key!
		pbEncoded = *data;
		memcpy(pbEncoded, pbPrivateKey, cbPrivateKey);
		return 0;
	}
    // now read the salt and number of rounds from the pbOptions!
    if (_libssh2_pem_getByteArrayOffset(&pbOptions,&cbOptions,&pbSalt,&cbSalt)!=0)    // salt
        return -1;
    rounds = _libssh2_pem_read32bit(&pbOptions,&cbOptions);
    // when we get here, we have the private key and encryption method -> decrypt it!
    strlwr(cipher);
    method=_libssh2_pem_findMethodByName(cipher);
    if (method==NULL)
        return -1;

    keyAndIV = (unsigned char*)malloc(method->secret_len+method->iv_len);

    ret=bcrypt_pbkdf(passphrase, strlen(passphrase), pbSalt, cbSalt, keyAndIV, method->secret_len+method->iv_len, rounds);
    if (ret==0) {
        ret=_libssh2_cipher_init(&ctx,method->algo,keyAndIV+method->secret_len,keyAndIV,0);
        if (ret==0) {
            pbData=pbPrivateKey;
            blockcount=cbPrivateKey/method->blocksize;
            for (blocknr=0;blocknr<blockcount;blocknr++) {
                _libssh2_cipher_crypt(&ctx,
                             method->algo,
                             0,
                             pbData,
                             method->blocksize);
                pbData+=method->blocksize;
            }
        }
    }
    memset(keyAndIV, 0, method->secret_len+method->iv_len);
    free(keyAndIV);
    testNr1=(unsigned int*)pbPrivateKey;
    testNr2=testNr1+1;
    if (ret || *testNr1!=*testNr2)
        return -1;
    pbEncoded = *data;
    memcpy(pbEncoded,pbPrivateKey,cbPrivateKey);
    return 0;
}

int
_libssh2_pem_get_public_key_bcrypt(LIBSSH2_SESSION *session,
                         unsigned char **pubkeydata,
                         size_t *pubkeydata_len,
                         unsigned char *pbEncoded,
                         unsigned int cbEncoded)

{
    const LIBSSH2_CRYPT_METHOD *method=NULL;
    unsigned char *cipherPtr,*pbOptions,*pbPublicKey;
    unsigned int cbOptions,cbPublicKey;
    unsigned int cipherLen,len;

    pbEncoded+=15;   // skip the code name "openssh-key-v1"
    cbEncoded-=15;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&cipherPtr,&cipherLen)!=0)
        return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&cipherPtr,&cipherLen)!=0)
        return -1;
    if (cipherLen!=6 || strncmp(cipherPtr,"bcrypt",6)!=0)
		if (cipherLen != 4 || strncmp(cipherPtr, "none", 4) != 0)
			return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&pbOptions,&cbOptions)!=0)
        return -1;
    len = _libssh2_pem_read32bit(&pbEncoded,&cbEncoded);
    if (len!=1)
        return -1;
    if (_libssh2_pem_getByteArrayOffset(&pbEncoded,&cbEncoded,&pbPublicKey,&cbPublicKey)!=0)    // public key
		return -1;
	*pubkeydata = LIBSSH2_ALLOC(session, cbPublicKey);
    memcpy(*pubkeydata, pbPublicKey, cbPublicKey);
	*pubkeydata_len = cbPublicKey;
    return 0;
}

